#!/usr/bin/env python
from __future__ import print_function

import fnmatch
import optparse
import os
import re
import sys
import textwrap
import time
import warnings

try:
    from paste.deploy import appconfig
except ImportError:
    appconfig = None

import sqlobject
from sqlobject import col
from sqlobject.classregistry import findClass
from sqlobject.declarative import DeclarativeMeta
from sqlobject.util import moduleloader
from sqlobject.compat import PY2, with_metaclass, string_type

# It's not very unsafe to use tempnam like we are doing:
warnings.filterwarnings(
    'ignore', 'tempnam is a potential security risk.*',
    RuntimeWarning, '.*command', 28)


if PY2:
    # noqa for flake8 and python 3
    input = raw_input  # noqa


def nowarning_tempnam(*args, **kw):
    return os.tempnam(*args, **kw)


class SQLObjectVersionTable(sqlobject.SQLObject):
    """
    This table is used to store information about the database and
    its version (used with record and update commands).
    """
    class sqlmeta:
        table = 'sqlobject_db_version'
    version = col.StringCol()
    updated = col.DateTimeCol(default=col.DateTimeCol.now)


def db_differences(soClass, conn):
    """
    Returns the differences between a class and the table in a
    connection.  Returns [] if no differences are found.  This
    function does the best it can; it can miss many differences.
    """
    # @@: Repeats a lot from CommandStatus.command, but it's hard
    # to actually factor out the display logic.  Or I'm too lazy
    # to do so.
    diffs = []
    if not conn.tableExists(soClass.sqlmeta.table):
        if soClass.sqlmeta.columns:
            diffs.append('Does not exist in database')
    else:
        try:
            columns = conn.columnsFromSchema(soClass.sqlmeta.table,
                                             soClass)
        except AttributeError:
            # Database does not support reading columns
            pass
        else:
            existing = {}
            for _col in columns:
                _col = _col.withClass(soClass)
                existing[_col.dbName] = _col
            missing = {}
            for _col in soClass.sqlmeta.columnList:
                if _col.dbName in existing:
                    del existing[_col.dbName]
                else:
                    missing[_col.dbName] = _col
            for _col in existing.values():
                diffs.append('Database has extra column: %s' % _col.dbName)
            for _col in missing.values():
                diffs.append('Database missing column: %s' % _col.dbName)
    return diffs


class CommandRunner(object):

    def __init__(self):
        self.commands = {}
        self.command_aliases = {}

    def run(self, argv):
        invoked_as = argv[0]
        args = argv[1:]
        for i in range(len(args)):
            if not args[i].startswith('-'):
                # this must be a command
                command = args[i].lower()
                del args[i]
                break
        else:
            # no command found
            self.invalid('No COMMAND given (try "%s help")'
                         % os.path.basename(invoked_as))
        real_command = self.command_aliases.get(command, command)
        if real_command not in self.commands.keys():
            self.invalid('COMMAND %s unknown' % command)
        runner = self.commands[real_command](
            invoked_as, command, args, self)
        runner.run()

    def register(self, command):
        name = command.name
        self.commands[name] = command
        for alias in command.aliases:
            self.command_aliases[alias] = name

    def invalid(self, msg, code=2):
        print(msg)
        sys.exit(code)

the_runner = CommandRunner()
register = the_runner.register


def standard_parser(connection=True, simulate=True,
                    interactive=False, find_modules=True):
    parser = optparse.OptionParser()
    parser.add_option('-v', '--verbose',
                      help='Be verbose (multiple times for more verbosity)',
                      action='count',
                      dest='verbose',
                      default=0)
    if simulate:
        parser.add_option('-n', '--simulate',
                          help="Don't actually do anything (implies -v)",
                          action='store_true',
                          dest='simulate')
    if connection:
        parser.add_option('-c', '--connection',
                          help="The database connection URI",
                          metavar='URI',
                          dest='connection_uri')
    parser.add_option('-f', '--config-file',
                      help="The Paste config file "
                      "that contains the database URI (in the database key)",
                      metavar="FILE",
                      dest="config_file")
    if find_modules:
        parser.add_option('-m', '--module',
                          help="Module in which to find SQLObject classes",
                          action='append',
                          metavar='MODULE',
                          dest='modules',
                          default=[])
        parser.add_option('-p', '--package',
                          help="Package to search for SQLObject classes",
                          action="append",
                          metavar="PACKAGE",
                          dest="packages",
                          default=[])
        parser.add_option('--class',
                          help="Select only named classes (wildcards allowed)",
                          action="append",
                          metavar="NAME",
                          dest="class_matchers",
                          default=[])
    if interactive:
        parser.add_option('-i', '--interactive',
                          help="Ask before doing anything "
                          "(use twice to be more careful)",
                          action="count",
                          dest="interactive",
                          default=0)
    parser.add_option('--egg',
                      help="Select modules from the given Egg, "
                      "using sqlobject.txt",
                      action="append",
                      metavar="EGG_SPEC",
                      dest="eggs",
                      default=[])
    return parser


class Command(with_metaclass(DeclarativeMeta, object)):

    min_args = 0
    min_args_error = 'You must provide at least %(min_args)s arguments'
    max_args = 0
    max_args_error = 'You must provide no more than %(max_args)s arguments'
    aliases = ()
    required_args = []
    description = None

    help = ''

    def orderClassesByDependencyLevel(self, classes):
        """
        Return classes ordered by their depth in the class dependency
        tree (this is *not* the inheritance tree), from the
        top level (independant) classes to the deepest level.
        The dependency tree is defined by the foreign key relations.
        """
        # @@: written as a self-contained function for now, to prevent
        # having to modify any core SQLObject component and namespace
        # contamination.
        # yemartin - 2006-08-08

        class SQLObjectCircularReferenceError(Exception):
            pass

        def findReverseDependencies(cls):
            """
            Return a list of classes that cls depends on. Note that
            "depends on" here mean "has a foreign key pointing to".
            """
            depended = []
            for _col in cls.sqlmeta.columnList:
                if _col.foreignKey:
                    other = findClass(_col.foreignKey,
                                      _col.soClass.sqlmeta.registry)
                    if (other is not cls) and (other not in depended):
                        depended.append(other)
            return depended

        # Cache to save already calculated dependency levels.
        dependency_levels = {}

        def calculateDependencyLevel(cls, dependency_stack=[]):
            """
            Recursively calculate the dependency level of cls, while
            using the dependency_stack to detect any circular reference.
            """
            # Return value from the cache if already calculated
            if cls in dependency_levels:
                return dependency_levels[cls]
            # Check for circular references
            if cls in dependency_stack:
                dependency_stack.append(cls)
                raise SQLObjectCircularReferenceError(
                    "Found a circular reference: %s " %
                    (' --> '.join([x.__name__ for x in dependency_stack])))
            dependency_stack.append(cls)
            # Recursively inspect dependent classes.
            depended = findReverseDependencies(cls)
            if depended:
                level = max([calculateDependencyLevel(x, dependency_stack)
                             for x in depended]) + 1
            else:
                level = 0
            dependency_levels[cls] = level
            return level

        # Now simply calculate and sort by dependency levels:
        try:
            sorter = []
            for cls in classes:
                level = calculateDependencyLevel(cls)
                sorter.append((level, cls))
            sorter.sort(key=lambda x: x[0])
            ordered_classes = [cls for _, cls in sorter]
        except SQLObjectCircularReferenceError as msg:
            # Failsafe: return the classes as-is if a circular reference
            # prevented the dependency levels to be calculated.
            print("Warning: a circular reference was detected in the "
                  "model. Unable to sort the classes by dependency: they "
                  "will be treated in alphabetic order. This may or may "
                  "not work depending on your database backend. "
                  "The error was:\n%s" % msg)
            return classes
        return ordered_classes

    def __classinit__(cls, new_args):
        if cls.__bases__ == (object,):
            # This abstract base class
            return
        register(cls)

    def __init__(self, invoked_as, command_name, args, runner):
        self.invoked_as = invoked_as
        self.command_name = command_name
        self.raw_args = args
        self.runner = runner

    def run(self):
        self.parser.usage = "%%prog [options]\n%s" % self.summary
        if self.help:
            help = textwrap.fill(
                self.help, int(os.environ.get('COLUMNS', 80)) - 4)
            self.parser.usage += '\n' + help
        self.parser.prog = '%s %s' % (
            os.path.basename(self.invoked_as),
            self.command_name)
        if self.description:
            self.parser.description = self.description
        self.options, self.args = self.parser.parse_args(self.raw_args)
        if (getattr(self.options, 'simulate', False)
                and not self.options.verbose):
            self.options.verbose = 1
        if self.min_args is not None and len(self.args) < self.min_args:
            self.runner.invalid(
                self.min_args_error % {'min_args': self.min_args,
                                       'actual_args': len(self.args)})
        if self.max_args is not None and len(self.args) > self.max_args:
            self.runner.invalid(
                self.max_args_error % {'max_args': self.max_args,
                                       'actual_args': len(self.args)})
        for var_name, option_name in self.required_args:
            if not getattr(self.options, var_name, None):
                self.runner.invalid(
                    'You must provide the option %s' % option_name)
        conf = self.config()
        if conf and conf.get('sys_path'):
            update_sys_path(conf['sys_path'], self.options.verbose)
        if conf and conf.get('database'):
            conn = sqlobject.connectionForURI(conf['database'])
            sqlobject.sqlhub.processConnection = conn
        for egg_spec in getattr(self.options, 'eggs', []):
            self.load_options_from_egg(egg_spec)
        self.command()

    def classes(self, require_connection=True,
                require_some=False):
        all = []
        for module_name in self.options.modules:
            all.extend(self.classes_from_module(
                moduleloader.load_module(module_name)))
        for package_name in self.options.packages:
            all.extend(self.classes_from_package(package_name))
        for egg_spec in self.options.eggs:
            all.extend(self.classes_from_egg(egg_spec))
        if self.options.class_matchers:
            filtered = []
            for soClass in all:
                name = soClass.__name__
                for matcher in self.options.class_matchers:
                    if fnmatch.fnmatch(name, matcher):
                        filtered.append(soClass)
                        break
            all = filtered
        conn = self.connection()
        if conn:
            for soClass in all:
                soClass._connection = conn
        else:
            missing = []
            for soClass in all:
                try:
                    if not soClass._connection:
                        missing.append(soClass)
                except AttributeError:
                    missing.append(soClass)
            if missing and require_connection:
                self.runner.invalid(
                    'These classes do not have connections set:\n  * %s\n'
                    'You must indicate --connection=URI'
                    % '\n  * '.join([soClass.__name__
                                     for soClass in missing]))
        if require_some and not all:
            print('No classes found!')
            if self.options.modules:
                print('Looked in modules: %s' %
                      ', '.join(self.options.modules))
            else:
                print('No modules specified')
            if self.options.packages:
                print('Looked in packages: %s' %
                      ', '.join(self.options.packages))
            else:
                print('No packages specified')
            if self.options.class_matchers:
                print('Matching class pattern: %s' %
                      self.options.class_matches)
            if self.options.eggs:
                print('Looked in eggs: %s' % ', '.join(self.options.eggs))
            else:
                print('No eggs specified')
            sys.exit(1)
        return self.orderClassesByDependencyLevel(all)

    def classes_from_module(self, module):
        all = []
        if hasattr(module, 'soClasses'):
            for name_or_class in module.soClasses:
                if isinstance(name_or_class, str):
                    name_or_class = getattr(module, name_or_class)
                all.append(name_or_class)
        else:
            for name in dir(module):
                value = getattr(module, name)
                if (isinstance(value, type)
                    and issubclass(value, sqlobject.SQLObject)
                        and value.__module__ == module.__name__):
                    all.append(value)
        return all

    def connection(self):
        config = self.config()
        if config is not None:
            assert config.get('database'), (
                "No database variable found in config file %s"
                % self.options.config_file)
            return sqlobject.connectionForURI(config['database'])
        elif getattr(self.options, 'connection_uri', None):
            return sqlobject.connectionForURI(self.options.connection_uri)
        else:
            return None

    def config(self):
        if not getattr(self.options, 'config_file', None):
            return None
        config_file = self.options.config_file
        if appconfig:
            if (not config_file.startswith('egg:')
                    and not config_file.startswith('config:')):
                config_file = 'config:' + config_file
            return appconfig(config_file,
                             relative_to=os.getcwd())
        else:
            return self.ini_config(config_file)

    def ini_config(self, conf_fn):
        conf_section = 'main'
        if '#' in conf_fn:
            conf_fn, conf_section = conf_fn.split('#', 1)

        try:
            from ConfigParser import ConfigParser
        except ImportError:
            from configparser import ConfigParser
        p = ConfigParser()
        # Case-sensitive:
        p.optionxform = str
        if not os.path.exists(conf_fn):
            # Stupid RawConfigParser doesn't give an error for
            # non-existant files:
            raise OSError(
                "Config file %s does not exist" % self.options.config_file)
        p.read([conf_fn])
        p._defaults.setdefault(
            'here', os.path.dirname(os.path.abspath(conf_fn)))

        possible_sections = []
        for section in p.sections():
            name = section.strip().lower()
            if (conf_section == name
                or (conf_section == name.split(':')[-1]
                    and name.split(':')[0] in ('app', 'application'))):
                possible_sections.append(section)

        if not possible_sections:
            raise OSError(
                "Config file %s does not have a section [%s] or [*:%s]"
                % (conf_fn, conf_section, conf_section))
        if len(possible_sections) > 1:
            raise OSError(
                "Config file %s has multiple sections matching %s: %s"
                % (conf_fn, conf_section, ', '.join(possible_sections)))

        config = {}
        for op in p.options(possible_sections[0]):
            config[op] = p.get(possible_sections[0], op)
        return config

    def classes_from_package(self, package_name):
        all = []
        package = moduleloader.load_module(package_name)
        package_dir = os.path.dirname(package.__file__)

        def find_classes_in_file(arg, dir_name, filenames):
            if dir_name.startswith('.svn'):
                return
            filenames = filter(
                lambda fname: fname.endswith('.py') and fname != '__init__.py',
                filenames)
            for fname in filenames:
                module_name = os.path.join(dir_name, fname)
                module_name = module_name[module_name.find(package_name):]
                module_name = module_name.replace(os.path.sep, '.')[:-3]
                try:
                    module = moduleloader.load_module(module_name)
                except ImportError as err:
                    if self.options.verbose:
                        print('Could not import module "%s". '
                              'Error was : "%s"' % (module_name, err))
                    continue
                except Exception as exc:
                    if self.options.verbose:
                        print('Unknown exception while processing module '
                              '"%s" : "%s"' % (module_name, exc))
                    continue
                classes = self.classes_from_module(module)
                all.extend(classes)

        for dirpath, dirnames, filenames in os.walk(package_dir):
            find_classes_in_file(None, dirpath, dirnames + filenames)
        return all

    def classes_from_egg(self, egg_spec):
        modules = []
        dist, conf = self.config_from_egg(egg_spec, warn_no_sqlobject=True)
        for mod in conf.get('db_module', '').split(','):
            mod = mod.strip()
            if not mod:
                continue
            if self.options.verbose:
                print('Looking in module %s' % mod)
            modules.extend(self.classes_from_module(
                moduleloader.load_module(mod)))
        return modules

    def load_options_from_egg(self, egg_spec):
        dist, conf = self.config_from_egg(egg_spec)
        if (hasattr(self.options, 'output_dir')
                and not self.options.output_dir and conf.get('history_dir')):
            dir = conf['history_dir']
            dir = dir.replace('$base', dist.location)
            self.options.output_dir = dir

    def config_from_egg(self, egg_spec, warn_no_sqlobject=True):
        import pkg_resources
        dist = pkg_resources.get_distribution(egg_spec)
        if not dist.has_metadata('sqlobject.txt'):
            if warn_no_sqlobject:
                print('No sqlobject.txt in %s egg info' % egg_spec)
            return None, {}
        result = {}
        for line in dist.get_metadata_lines('sqlobject.txt'):
            line = line.strip()
            if not line or line.startswith('#'):
                continue
            name, value = line.split('=', 1)
            name = name.strip().lower()
            if name in result:
                print('Warning: %s appears more than once '
                      'in sqlobject.txt' % name)
            result[name.strip().lower()] = value.strip()
        return dist, result

    def command(self):
        raise NotImplementedError

    def _get_prog_name(self):
        return os.path.basename(self.invoked_as)
    prog_name = property(_get_prog_name)

    def ask(self, prompt, safe=False, default=True):
        if self.options.interactive >= 2:
            default = safe
        if default:
            prompt += ' [Y/n]? '
        else:
            prompt += ' [y/N]? '
        while 1:
            response = input(prompt).strip()
            if not response.strip():
                return default
            if response and response[0].lower() in ('y', 'n'):
                return response[0].lower() == 'y'
            print('Y or N please')

    def shorten_filename(self, fn):
        """
        Shortens a filename to make it relative to the current
        directory (if it can).  For display purposes.
        """
        if fn.startswith(os.getcwd() + '/'):
            fn = fn[len(os.getcwd()) + 1:]
        return fn

    def open_editor(self, pretext, breaker=None, extension='.txt'):
        """
        Open an editor with the given text.  Return the new text,
        or None if no edits were made.  If given, everything after
        `breaker` will be ignored.
        """
        fn = nowarning_tempnam() + extension
        f = open(fn, 'w')
        f.write(pretext)
        f.close()
        print('$EDITOR %s' % fn)
        os.system('$EDITOR %s' % fn)
        f = open(fn, 'r')
        content = f.read()
        f.close()
        if breaker:
            content = content.split(breaker)[0]
            pretext = pretext.split(breaker)[0]
        if content == pretext or not content.strip():
            return None
        return content


class CommandSQL(Command):

    name = 'sql'
    summary = 'Show SQL CREATE statements'

    parser = standard_parser(simulate=False)

    def command(self):
        classes = self.classes()
        allConstraints = []
        for cls in classes:
            if self.options.verbose >= 1:
                print('-- %s from %s' % (
                      cls.__name__, cls.__module__))
            createSql, constraints = cls.createTableSQL()
            print(createSql.strip() + ';\n')
            allConstraints.append(constraints)
        for constraints in allConstraints:
            if constraints:
                for constraint in constraints:
                    if constraint:
                        print(constraint.strip() + ';\n')


class CommandList(Command):

    name = 'list'
    summary = 'Show all SQLObject classes found'

    parser = standard_parser(simulate=False, connection=False)

    def command(self):
        if self.options.verbose >= 1:
            print('Classes found:')
        classes = self.classes(require_connection=False)
        for soClass in classes:
            print('%s.%s' % (soClass.__module__, soClass.__name__))
            if self.options.verbose >= 1:
                print('  Table: %s' % soClass.sqlmeta.table)


class CommandCreate(Command):

    name = 'create'
    summary = 'Create tables'

    parser = standard_parser(interactive=True)
    parser.add_option('--create-db',
                      action='store_true',
                      dest='create_db',
                      help="Create the database")

    def command(self):
        v = self.options.verbose
        created = 0
        existing = 0
        dbs_created = []
        constraints = {}
        for soClass in self.classes(require_some=True):
            if (self.options.create_db
                    and soClass._connection not in dbs_created):
                if not self.options.simulate:
                    try:
                        soClass._connection.createEmptyDatabase()
                    except soClass._connection.module.ProgrammingError as e:
                        if str(e).find('already exists') != -1:
                            print('Database already exists')
                        else:
                            raise
                else:
                    print('(simulating; cannot create database)')
                dbs_created.append(soClass._connection)
            if soClass._connection not in constraints.keys():
                constraints[soClass._connection] = []
            exists = soClass._connection.tableExists(soClass.sqlmeta.table)
            if v >= 1:
                if exists:
                    existing += 1
                    print('%s already exists.' % soClass.__name__)
                else:
                    print('Creating %s' % soClass.__name__)
            if v >= 2:
                sql, extra = soClass.createTableSQL()
                print(sql)
            if (not self.options.simulate and not exists):
                if self.options.interactive:
                    if self.ask('Create %s' % soClass.__name__):
                        created += 1
                        tableConstraints = soClass.createTable(
                            applyConstraints=False)
                        if tableConstraints:
                            constraints[soClass._connection].append(
                                tableConstraints)
                    else:
                        print('Cancelled')
                else:
                    created += 1
                    tableConstraints = soClass.createTable(
                        applyConstraints=False)
                    if tableConstraints:
                        constraints[soClass._connection].append(
                            tableConstraints)
        for connection in constraints.keys():
            if v >= 2:
                print('Creating constraints')
            for constraintList in constraints[connection]:
                for constraint in constraintList:
                    if constraint:
                        connection.query(constraint)
        if v >= 1:
            print('%i tables created (%i already exist)' % (
                  created, existing))


class CommandDrop(Command):

    name = 'drop'
    summary = 'Drop tables'

    parser = standard_parser(interactive=True)

    def command(self):
        v = self.options.verbose
        dropped = 0
        not_existing = 0
        for soClass in reversed(self.classes()):
            exists = soClass._connection.tableExists(soClass.sqlmeta.table)
            if v >= 1:
                if exists:
                    print('Dropping %s' % soClass.__name__)
                else:
                    not_existing += 1
                    print('%s does not exist.' % soClass.__name__)
            if (not self.options.simulate and exists):
                if self.options.interactive:
                    if self.ask('Drop %s' % soClass.__name__):
                        dropped += 1
                        soClass.dropTable()
                    else:
                        print('Cancelled')
                else:
                    dropped += 1
                    soClass.dropTable()
        if v >= 1:
            print('%i tables dropped (%i didn\'t exist)' % (
                  dropped, not_existing))


class CommandStatus(Command):

    name = 'status'
    summary = 'Show status of classes vs. database'
    help = ('This command checks the SQLObject definition and checks if '
            'the tables in the database match.  It can always test for '
            'missing tables, and on some databases can test for the '
            'existance of other tables.  Column types are not currently '
            'checked.')

    parser = standard_parser(simulate=False)

    def print_class(self, soClass):
        if self.printed:
            return
        self.printed = True
        print('Checking %s...' % soClass.__name__)

    def command(self):
        good = 0
        bad = 0
        missing_tables = 0
        columnsFromSchema_warning = False
        for soClass in self.classes(require_some=True):
            conn = soClass._connection
            self.printed = False
            if self.options.verbose:
                self.print_class(soClass)
            if not conn.tableExists(soClass.sqlmeta.table):
                self.print_class(soClass)
                print('  Does not exist in database')
                missing_tables += 1
                continue
            try:
                columns = conn.columnsFromSchema(soClass.sqlmeta.table,
                                                 soClass)
            except AttributeError:
                if not columnsFromSchema_warning:
                    print('Database does not support reading columns')
                    columnsFromSchema_warning = True
                good += 1
                continue
            except AssertionError as e:
                print('Cannot read db table %s: %s' % (
                    soClass.sqlmeta.table, e))
                continue
            existing = {}
            for _col in columns:
                _col = _col.withClass(soClass)
                existing[_col.dbName] = _col
            missing = {}
            for _col in soClass.sqlmeta.columnList:
                if _col.dbName in existing:
                    del existing[_col.dbName]
                else:
                    missing[_col.dbName] = _col
            if existing:
                self.print_class(soClass)
                for _col in existing.values():
                    print('  Database has extra column: %s' % _col.dbName)
            if missing:
                self.print_class(soClass)
                for _col in missing.values():
                    print('  Database missing column: %s' % _col.dbName)
            if existing or missing:
                bad += 1
            else:
                good += 1
        if self.options.verbose:
            print('%i in sync; %i out of sync; %i not in database' % (
                  good, bad, missing_tables))


class CommandHelp(Command):

    name = 'help'
    summary = 'Show help'

    parser = optparse.OptionParser()

    max_args = 1

    def command(self):
        if self.args:
            the_runner.run([self.invoked_as, self.args[0], '-h'])
        else:
            print('Available commands:')
            print('  (use "%s help COMMAND" or "%s COMMAND -h" ' % (
                  self.prog_name, self.prog_name))
            print('  for more information)')
            items = sorted(the_runner.commands.items())
            max_len = max([len(cn) for cn, c in items])
            for command_name, command in items:
                print('%s:%s %s' % (command_name,
                                    ' ' * (max_len - len(command_name)),
                                    command.summary))
                if command.aliases:
                    print('%s (Aliases: %s)' % (
                        ' ' * max_len, ', '.join(command.aliases)))


class CommandExecute(Command):

    name = 'execute'
    summary = 'Execute SQL statements'
    help = ('Runs SQL statements directly in the database, with no '
            'intervention.  Useful when used with a configuration file.  '
            'Each argument is executed as an individual statement.')

    parser = standard_parser(find_modules=False)
    parser.add_option('--stdin',
                      help="Read SQL from stdin "
                      "(normally takes SQL from the command line)",
                      dest="use_stdin",
                      action="store_true")

    max_args = None

    def command(self):
        args = self.args
        if self.options.use_stdin:
            if self.options.verbose:
                print("Reading additional SQL from stdin "
                      "(Ctrl-D or Ctrl-Z to finish)...")
            args.append(sys.stdin.read())
        self.conn = self.connection().getConnection()
        self.cursor = self.conn.cursor()
        for sql in args:
            self.execute_sql(sql)

    def execute_sql(self, sql):
        if self.options.verbose:
            print(sql)
        try:
            self.cursor.execute(sql)
        except Exception as e:
            if not self.options.verbose:
                print(sql)
            print("****Error:")
            print('    ', e)
            return
        desc = self.cursor.description
        rows = self.cursor.fetchall()
        if self.options.verbose:
            if not self.cursor.rowcount:
                print("No rows accessed")
            else:
                print("%i rows accessed" % self.cursor.rowcount)
        if desc:
            for (name, type_code, display_size, internal_size,
                    precision, scale, null_ok) in desc:
                sys.stdout.write("%s\t" % name)
            sys.stdout.write("\n")
        for row in rows:
            for _col in row:
                sys.stdout.write("%r\t" % _col)
            sys.stdout.write("\n")
        print()


class CommandRecord(Command):

    name = 'record'
    summary = 'Record historical information about the database status'
    help = ('Record state of table definitions.  The state of each '
            'table is written out to a separate file in a directory, '
            'and that directory forms a "version".  A table is also '
            'added to your database (%s) that reflects the version the '
            'database is currently at.  Use the upgrade command to '
            'sync databases with code.'
            % SQLObjectVersionTable.sqlmeta.table)

    parser = standard_parser()
    parser.add_option('--output-dir',
                      help="Base directory for recorded definitions",
                      dest="output_dir",
                      metavar="DIR",
                      default=None)
    parser.add_option('--no-db-record',
                      help="Don't record version to database",
                      dest="db_record",
                      action="store_false",
                      default=True)
    parser.add_option('--force-create',
                      help="Create a new version even if appears to be "
                      "identical to the last version",
                      action="store_true",
                      dest="force_create")
    parser.add_option('--name',
                      help="The name to append to the version.  The "
                      "version should sort after previous versions (so "
                      "any versions from the same day should come "
                      "alphabetically before this version).",
                      dest="version_name",
                      metavar="NAME")
    parser.add_option('--force-db-version',
                      help="Update the database version, and include no "
                      "database information.  This is for databases that "
                      "were developed without any interaction with "
                      "this tool, to create a 'beginning' revision.",
                      metavar="VERSION_NAME",
                      dest="force_db_version")
    parser.add_option('--edit',
                      help="Open an editor for the upgrader in the last "
                      "version (using $EDITOR).",
                      action="store_true",
                      dest="open_editor")

    version_regex = re.compile(r'^\d\d\d\d-\d\d-\d\d')

    def command(self):
        if self.options.force_db_version:
            self.command_force_db_version()
            return

        v = self.options.verbose
        sim = self.options.simulate
        classes = self.classes()
        if not classes:
            print("No classes found!")
            return

        output_dir = self.find_output_dir()
        version = os.path.basename(output_dir)
        print("Creating version %s" % version)
        conns = []
        files = {}
        for cls in self.classes():
            dbName = cls._connection.dbName
            if cls._connection not in conns:
                conns.append(cls._connection)
            fn = os.path.join(cls.__name__ + '_' + dbName + '.sql')
            if sim:
                continue
            create, constraints = cls.createTableSQL()
            if constraints:
                constraints = '\n-- Constraints:\n%s\n' % (
                    '\n'.join(constraints))
            else:
                constraints = ''
            files[fn] = ''.join([
                '-- Exported definition from %s\n'
                % time.strftime('%Y-%m-%dT%H:%M:%S'),
                '-- Class %s.%s\n'
                % (cls.__module__, cls.__name__),
                '-- Database: %s\n'
                % dbName,
                create.strip(),
                '\n',
                constraints])
        last_version_dir = self.find_last_version()
        if last_version_dir and not self.options.force_create:
            if v > 1:
                print("Checking %s to see if it is current" % last_version_dir)
            files_copy = files.copy()
            for fn in os.listdir(last_version_dir):
                if not fn.endswith('.sql'):
                    continue
                if fn not in files_copy:
                    if v > 1:
                        print("Missing file %s" % fn)
                    break
                f = open(os.path.join(last_version_dir, fn), 'r')
                content = f.read()
                f.close()
                if (self.strip_comments(files_copy[fn])
                        != self.strip_comments(content)):
                    if v > 1:
                        print("Content does not match: %s" % fn)
                    break
                del files_copy[fn]
            else:
                # No differences so far
                if not files_copy:
                    # Used up all files
                    print("Current status matches version %s"
                          % os.path.basename(last_version_dir))
                    return
                if v > 1:
                    print("Extra files: %s" % ', '.join(files_copy.keys()))
            if v:
                print("Current state does not match %s"
                      % os.path.basename(last_version_dir))
        if v > 1 and not last_version_dir:
            print("No last version to check")
        if not sim:
            os.mkdir(output_dir)
        if v:
            print('Making directory %s' % self.shorten_filename(output_dir))
        files = sorted(files.items())
        for fn, content in files:
            if v:
                print('  Writing %s' % self.shorten_filename(fn))
            if not sim:
                f = open(os.path.join(output_dir, fn), 'w')
                f.write(content)
                f.close()
        if self.options.db_record:
            all_diffs = []
            for cls in self.classes():
                for conn in conns:
                    diffs = db_differences(cls, conn)
                    for diff in diffs:
                        if len(conns) > 1:
                            diff = '  (%s).%s: %s' % (
                                conn.uri(), cls.sqlmeta.table, diff)
                        else:
                            diff = '  %s: %s' % (cls.sqlmeta.table, diff)
                        all_diffs.append(diff)
            if all_diffs:
                print('Database does not match schema:')
                print('\n'.join(all_diffs))
                for conn in conns:
                    self.update_db(version, conn)
        else:
            all_diffs = []
        if self.options.open_editor:
            if not last_version_dir:
                print("Cannot edit upgrader because there is no "
                      "previous version")
            else:
                breaker = ('-' * 20 + ' lines below this will be ignored '
                           + '-' * 20)
                pre_text = breaker + '\n' + '\n'.join(all_diffs)
                text = self.open_editor('\n\n' + pre_text, breaker=breaker,
                                        extension='.sql')
                if text is not None:
                    fn = os.path.join(last_version_dir,
                                      'upgrade_%s_%s.sql' %
                                      (dbName, version))
                    f = open(fn, 'w')
                    f.write(text)
                    f.close()
                    print('Wrote to %s' % fn)

    def update_db(self, version, conn):
        v = self.options.verbose
        if not conn.tableExists(SQLObjectVersionTable.sqlmeta.table):
            if v:
                print('Creating table %s'
                      % SQLObjectVersionTable.sqlmeta.table)
            sql = SQLObjectVersionTable.createTableSQL(connection=conn)
            if v > 1:
                print(sql)
            if not self.options.simulate:
                SQLObjectVersionTable.createTable(connection=conn)
        if not self.options.simulate:
            SQLObjectVersionTable.clearTable(connection=conn)
            SQLObjectVersionTable(
                version=version,
                connection=conn)

    def strip_comments(self, sql):
        lines = [l for l in sql.splitlines()
                 if not l.strip().startswith('--')]
        return '\n'.join(lines)

    def base_dir(self):
        base = self.options.output_dir
        if base is None:
            config = self.config()
            if config is not None:
                base = config.get('sqlobject_history_dir', '.')
            else:
                base = '.'
        if not os.path.exists(base):
            print('Creating history directory %s' %
                  self.shorten_filename(base))
            if not self.options.simulate:
                os.makedirs(base)
        return base

    def find_output_dir(self):
        today = time.strftime('%Y-%m-%d', time.localtime())
        if self.options.version_name:
            dir = os.path.join(self.base_dir(), today + '-'
                               + self.options.version_name)
            if os.path.exists(dir):
                print("Error, directory already exists: %s"
                      % dir)
                sys.exit(1)
            return dir
        extra = ''
        while 1:
            dir = os.path.join(self.base_dir(), today + extra)
            if not os.path.exists(dir):
                return dir
            if not extra:
                extra = 'a'
            else:
                extra = chr(ord(extra) + 1)

    def find_last_version(self):
        names = []
        for fn in os.listdir(self.base_dir()):
            if not self.version_regex.search(fn):
                continue
            names.append(fn)
        if not names:
            return None
        names.sort()
        return os.path.join(self.base_dir(), names[-1])

    def command_force_db_version(self):
        v = self.options.verbose
        sim = self.options.simulate
        version = self.options.force_db_version
        if not self.version_regex.search(version):
            print("Versions must be in the format YYYY-MM-DD...")
            print("You version %s does not fit this" % version)
            return
        version_dir = os.path.join(self.base_dir(), version)
        if not os.path.exists(version_dir):
            if v:
                print('Creating %s' % self.shorten_filename(version_dir))
            if not sim:
                os.mkdir(version_dir)
        elif v:
            print('Directory %s exists'
                  % self.shorten_filename(version_dir))
        if self.options.db_record:
            self.update_db(version, self.connection())


class CommandUpgrade(CommandRecord):

    name = 'upgrade'
    summary = 'Update the database to a new version (as created by record)'
    help = ('This command runs scripts (that you write by hand) to '
            'upgrade a database.  The database\'s current version is in '
            'the sqlobject_version table (use record --force-db-version '
            'if a database does not have a sqlobject_version table), '
            'and upgrade scripts are in the version directory you are '
            'upgrading FROM, named upgrade_DBNAME_VERSION.sql, like '
            '"upgrade_mysql_2004-12-01b.sql".')

    parser = standard_parser(find_modules=False)
    parser.add_option('--upgrade-to',
                      help="Upgrade to the given version "
                      "(default: newest version)",
                      dest="upgrade_to",
                      metavar="VERSION")
    parser.add_option('--output-dir',
                      help="Base directory for recorded definitions",
                      dest="output_dir",
                      metavar="DIR",
                      default=None)

    upgrade_regex = re.compile(r'^upgrade_([a-z]*)_([^.]*)\.sql$', re.I)

    def command(self):
        v = self.options.verbose
        sim = self.options.simulate
        if self.options.upgrade_to:
            version_to = self.options.upgrade_to
        else:
            fname = self.find_last_version()
            if fname is None:
                print("No version exists, use 'record' command to create one")
                return
            version_to = os.path.basename(fname)
        current = self.current_version()
        if v:
            print('Current version: %s' % current)
        version_list = self.make_plan(current, version_to)
        if not version_list:
            print('Database up to date')
            return
        if v:
            print('Plan:')
            for next_version, upgrader in version_list:
                print('  Use %s to upgrade to %s' % (
                      self.shorten_filename(upgrader), next_version))
        conn = self.connection()
        for next_version, upgrader in version_list:
            f = open(upgrader)
            sql = f.read()
            f.close()
            if v:
                print("Running:")
                print(sql)
                print('-' * 60)
            if not sim:
                try:
                    conn.query(sql)
                except Exception:
                    print("Error in script: %s" % upgrader)
                    raise
            self.update_db(next_version, conn)
        print('Done.')

    def current_version(self):
        conn = self.connection()
        if not conn.tableExists(SQLObjectVersionTable.sqlmeta.table):
            print('No sqlobject_version table!')
            sys.exit(1)
        versions = list(SQLObjectVersionTable.select(connection=conn))
        if not versions:
            print('No rows in sqlobject_version!')
            sys.exit(1)
        if len(versions) > 1:
            print('Ambiguous sqlobject_version_table')
            sys.exit(1)
        return versions[0].version

    def make_plan(self, current, dest):
        if current == dest:
            return []
        dbname = self.connection().dbName
        next_version, upgrader = self.best_upgrade(current, dest, dbname)
        if not upgrader:
            print('No way to upgrade from %s to %s' % (current, dest))
            print('(you need a %s/upgrade_%s_%s.sql script)'
                  % (current, dbname, dest))
            sys.exit(1)
        plan = [(next_version, upgrader)]
        if next_version == dest:
            return plan
        else:
            return plan + self.make_plan(next_version, dest)

    def best_upgrade(self, current, dest, target_dbname):
        current_dir = os.path.join(self.base_dir(), current)
        if self.options.verbose > 1:
            print('Looking in %s for upgraders'
                  % self.shorten_filename(current_dir))
        upgraders = []
        for fn in os.listdir(current_dir):
            match = self.upgrade_regex.search(fn)
            if not match:
                if self.options.verbose > 1:
                    print('Not an upgrade script: %s' % fn)
                continue
            dbname = match.group(1)
            version = match.group(2)
            if dbname != target_dbname:
                if self.options.verbose > 1:
                    print('Not for this database: %s (want %s)' % (
                          dbname, target_dbname))
                continue
            if version > dest:
                if self.options.verbose > 1:
                    print('Version too new: %s (only want %s)' % (
                          version, dest))
            upgraders.append((version, os.path.join(current_dir, fn)))
        if not upgraders:
            if self.options.verbose > 1:
                print('No upgraders found in %s' % current_dir)
            return None, None
        upgraders.sort()
        return upgraders[-1]


def update_sys_path(paths, verbose):
    if isinstance(paths, string_type):
        paths = [paths]
    for path in paths:
        path = os.path.abspath(path)
        if path not in sys.path:
            if verbose > 1:
                print('Adding %s to path' % path)
            sys.path.insert(0, path)

if __name__ == '__main__':
    the_runner.run(sys.argv)



#!/usr/bin/python2.2

from slides import Lecture, NumSlide, Slide, Bullet, SubBullet, PRE, URL

lecture = Lecture(
	"High-Level Database Interaction with SQLObject",

	Slide("What is SQLObject?",
		  Bullet("YAORM -- Yet Another Object-Relational Mapper"),
		  Bullet("Allows you to map your Python classes to your database schema"),
		  Bullet("Takes the 'SQL' out of 'Database Programming'"),
		  Bullet("No special XML files to create, just normal Python classes")),
	Slide("Why Do I Care?",
		  Bullet("Storing things in databases is fairly common in day-to-day programming"),
		  Bullet("SQL is the standard language used to manipulate data in a database"),
		  Bullet("Writing SQL is boring, repetitive and depressing"),
		  Bullet("SQLObject relieves you of the burden of writing SQL"),
		  Bullet("...but still lets you write SQL when you need to")),
	Slide("How does SQLObject differ from other ORM's?",
		  Bullet("Simple is better than complex; SQLObject is very simple to use"),
		  Bullet("Mappings are defined using normal Python classes"),
		  Bullet("Uses properties instead of wordy get/set methods for column attributes"),
		  Bullet("No awkward auto-generation of Python classes/files from an external format"),
		  Bullet("Can generate the database schema directly from your Python classes, and vice versa"),
		  Bullet("Overloading magic provides convenient high-level SQL conditionals (dot q magic, stay tuned)")),
	Slide("A Simple Example",
		  Bullet("To create a table that looks like this in SQL:",
				 PRE("""\
CREATE TABLE person (
    id SERIAL,
    first_name VARCHAR(100) NOT NULL,
    middle_initial CHAR(1),
    last_name VARCHAR(150) NOT NULL
);""")),
		  Bullet("You would write this Python code:",
				 PRE("""\
class Person(SQLObject):
    firstName = StringCol(length = 100)
    middleInitial = StringCol(length = 1)
    lastName = StringCol(length = 150)""")),
		  Bullet("No, you're not dreaming :P")),
	Slide("Declaring the Class",
		  PRE("""\
from SQLObject import *

conn = PostgresConnection(db = 'testdb', user = 'testuser', password = 'testpass')

class Person(SQLObject):
    _connection = conn
	
    firstName = StringCol(length = 100)
    middleInitial = StringCol(length = 1)
    lastName = StringCol(length = 150)"""),
		  Bullet("Use one of MySQLConnection, PostgresConnection, SQLiteConnection or DBMConnection as your _connection"),

		  Bullet("Use StudlyCaps for your classes and mixedCase for your columns"),
		  Bullet("SQLObject will map TableName to table_name and columnName to column_name"),
	      Bullet("In the above example, class Person becomes table person with columns first_name, middle_initial and last_name")),
	Slide("Creating and Dropping Tables",
		  Bullet("Use the createTable class method to create the table, and two optional keyword arguments",SubBullet(
	Bullet("ifNotExists: only try to create the table if it doesn't exist"),
	Bullet("createJoinTables: will create the intermediate tables for many-to-many relationships"))),
		  Bullet("Conversely, use dropTable, passing in optional ifExists and dropJoinTables keyword arguments")),

	Slide("More on Column Syntax",
		  Bullet("Columns are implemented as properties"),
		  Bullet("Columns supported: StringCol, IntCol, FloatCol, EnumCol, DateTimeCol, ForeignKey, DecimalCol, CurrencyCol"),
		  Bullet("The first argument is the column name"),
		  Bullet("Keyword arguments specify additional information (e.g. notNone, default, length, etc)"),
		  Bullet("SQLObject lets the database do type checking and coercion"),
		  Bullet("An id column is implicitly created")),
	Slide("Keyword Arguments for Col Classes",
		  Bullet("dbName: the column name in the database"),
		  Bullet("default: the default value (can be a callable that returns a value)"),
		  Bullet("alternateID: set this to True if you want a byColumnName method to lookup rows based on this column"),
		  Bullet("unique: declare this column as UNIQUE in the database"),
		  Bullet("notNone: when True, column cannot be None/NULL"),
		  Bullet("sqlType: to specify the column type manually")),
	Slide("Special Class Attributes",
		  Bullet("_connection: the database connection for this class"),
		  Bullet("_table: the database name of the table behind this class"),
		  Bullet("_joins: a list of join relationships to other classes"),
		  Bullet("_cacheValues: you'll want to set this false if you're using SQLObject classes from multiple processes"),
		  Bullet("_idName: the name of the PK (defaults to id)"),
		  Bullet("_style: a style object that provides a custom Python to DB name mapping algorithm")),
	Slide("Using SQLObject classes",
		  Bullet("Create a new row with the .new() class method:",
				 PRE("""Person.new(firstName = "Brad", middleInitial = "E", lastName = "Bollenbach")""")),
		  Bullet("The row is inserted into the database as soon as you call .new()"),
		  Bullet("Access an existing row by passing an ID to the class constructor",
				 PRE("""\
>>> me = Person(1)
>>> me.firstName
'Brad'
>>> me.lastName
'Bollenbach'""")),
		  Bullet("Modify column values by modifying property values"),
		  Bullet("Changes to your object's properties are updated immediately in the database"),
		  Bullet("...but transactions are there if you need them (as long as the database supports them)")),
	Slide("Relating Your Classes with Joins",
		  Bullet("Use a ForeignKey to point a column's value at an instance of another class"),
		  Bullet("To relate the PK class back to the FK class, use MultipleJoin"),
		  Bullet("Let's give a Person some PhoneNumbers:",
				 PRE("""\
from SQLObject import *
conn = PostgresConnection(db = 'testdb', user = 'testuser', password = 'testpass')

class Person(SQLObject):
    _connection = conn
    
    firstName = StringCol(length = 100)
    middleInitial = StringCol(length = 1)
    lastName = StringCol(length = 150)
    phoneNumbers = MultipleJoin("PhoneNumber") 

class PhoneNumber(SQLObject):
    _connection = conn

    person = ForeignKey('Person')
    phoneNumber = StringCol(length = 10)"""))),
	Slide("Many-to-Many Relationships",
		  Bullet("A Person might have many Roles"),
		  Bullet("A Role might be associated to more than one Person"),
		  Bullet("Use a RelatedJoin to specify this many-to-many relation:",
				 PRE("""\
class Role(SQLObject):
    _connection = conn

    name = StringCol(length = 20)
    people = RelatedJoin('Person')

class Person(SQLObject):
    _connection = conn

    firstName = StringCol(length = 100)
    middleInitial = StringCol(length = 1)
    lastName = StringCol(length = 150)
    phoneNumbers = MultipleJoin("PhoneNumber")

    roles = RelatedJoin('Role')
	
me = Person.new(firstName = "Brad", middleInitial = "E", lastName = "Bollenbach")
pg = Role.new(name = "Python Geek")
me.addRole(pg)
""")),
		  Bullet("SQLObject added .addRole() and .removeRole() methods to Person, and .addPerson() and .removePerson() methods to Role"),
		  Bullet("...and created the person_role table by combining the table names of the classes alphabetically")),
	Slide("Selecting Multiple Objects",
		  Bullet("The class's .select() method can be used to return multiple objects of a given class"),
		  Bullet("Here comes that dot q magic!"),
		  Bullet("Every SQLObject derived class has a special .q attribute, which uses some funky operator overloading to construct queries using Python:",
				 PRE("""\
>>> from person import Person
>>> brads = Person.select(Person.q.firstName == 'Brad')
>>> list(brads)
[<Person 1 lastName='Bollenbach' middleInitial='E' firstName='Brad'>]
>>> brads[0].lastName
'Bollenbach'
""")),
		  Bullet("select accepts orderBy and groupBy arguments, which are strings (or lists of strings) referring to the database column name")),
	Slide("Selects using AND",
		  Bullet("Use the AND function for more specific queries:",
				 PRE("""\
>>> from SQLObject import AND
>>> from person import Person
>>> bradbs = Person.select(AND(
...     Person.q.firstName == "Brad",
...     Person.q.lastName.startswith("B")
... ))
""")),
		  Bullet("Notice that columns-as-properties behave much like their regular Python brethren (e.g. the startswith method used above)")),
	Slide("Customizing Column Behaviour",
		  Bullet("SQLObject uses metaclasses to generate the classes that map to your database tables"),
		  Bullet("Because of this, you need to take care when overriding property get/set methods"),
		  Bullet("You cannot simply override the get or set method and then call the parent's get/set"),
		  Bullet("Your class was created on-the-fly by SQLObject, so there is no parent! :)"),
		  Bullet("SQLObject gives you special _SO_set_foo and _SO_get_foo methods to call after you've done something special in the normal setter/getter"),
		  Bullet("An (admittedly contrived) example:",
				 PRE("""\
class Person(SQLObject):
    ...
    def _set_lastName(self, value):
        print "you changed the lastName"
        self._SO_set_lastName(value)
"""))),
	Slide("Better Living Through Transactions",
		  Bullet("Sometimes you need indivisibility"),
		  Bullet("SQLObject lets you use transactions when the database supports it",
				 PRE("""\
trans = Person._connection.transaction()
me = Person(1, trans)
me.firstName = "Bradley"
trans.commit()
me.firstName = "Yada"
trans.rollback()"""))),
	Slide("Automatic Class Generation",
		  Bullet("Laziness is a virtue"),
		  Bullet("If you've already got a database schema, why not make SQLObject create the classes for you?",
				 PRE("""\
class Person(SQLObject):

    _fromDatabase = True""")),
		  Bullet("You can still specify additional columns, as normal"),
		  Bullet("Currently this only works with MySQL")),
	Slide("SQLObject in the Wild -- Who's Using It?",
		  Bullet("BBnet.ca used SQLObject to develop a time tracking and invoicing system for a consultancy firm"),
		  Bullet("XSOLI (my current employer) is using SQLObject to drive an online merchant proxy system"),
		  Bullet("Future projects will probably include porting a sales application for a fixed-rate long distance provider onto the SQLObject framework"),
		  Bullet("There's an active community of SQLObject developers on the mailing list")),
	Slide("Future Plans for SQLObject",
		  Bullet("Add support for more databases"),
		  Bullet("Improved transaction handling, so that even non-transaction aware backends can support them"),
		  Bullet("Hooks into a validation and form generation package"),
		  Bullet("More powerful joins")),
	Slide("Summing Up",
		  Bullet("SQLObject is a framework that can make your (programming) life easier"),
		  Bullet("It provides a high-level Python interface to things you'd normally suffer through in SQL"),
		  Bullet("It offers advanced features like automatic class generation, dot q magic, and various ways of relating your classes"),
		  Bullet("There's an active developer base, so you're sure to find help when you need it")),
	Slide("Resources",
		  Bullet("SQLObject homepage -- ", URL("http://www.sqlobject.org")),
		  Bullet("Join the mailing list -- ", URL("http://lists.sourceforge.net/lists/listinfo/sqlobject-discuss")),
		  Bullet("Contact me -- brad@bbnet.ca")),
	Slide("Questions?",
		  Bullet("Questions?"),
		  Bullet("Comments?"),
		  Bullet("Free beer?")))

lecture.renderHTML(".", "slide-%02d.html", css="main.css")



import pytest
from sqlobject import SQLObject, JsonbCol
from sqlobject.tests.dbtest import getConnection, setupClass


########################################
# Jsonb columns
########################################


class JsonbContainer(SQLObject):
    jsondata = JsonbCol(default=None)
    jsondata_none = JsonbCol(default=None)
    jsondata_list = JsonbCol(default=None)


dictdata = dict(
    ship=u'USS Voyager',
    number='NCC-74656',
    crew=['Cpt. Janeway', 'Doctor', 'Lt. Torries', 'Seven of Nine'],
    races=('Borg', 'Kason',),
    distance2earth=70000,
    coming_home=True)

nonedata = None

listofdictsdata = [dict(Erath=7390000000),
                   dict(Vulcan=6000000000),
                   dict(Kronos='unknown')]


def test_jsonbCol():
    connection = getConnection()
    if connection.dbName != "postgres":
        pytest.skip("These tests require PostgreSQL")

    setupClass([JsonbContainer], force=True)

    my_jsonb = JsonbContainer(
        jsondata=dictdata,
        jsondata_none=nonedata,
        jsondata_list=listofdictsdata)
    iid = my_jsonb.id

    JsonbContainer._connection.cache.clear()

    my_jsonb_2 = JsonbContainer.get(iid)

    assert my_jsonb_2.jsondata['coming_home'] == dictdata['coming_home']
    assert my_jsonb_2.jsondata['crew'] == dictdata['crew']
    assert my_jsonb_2.jsondata['races'] == ['Borg', 'Kason']
    assert my_jsonb_2.jsondata_none is None
    assert my_jsonb_2.jsondata_list == listofdictsdata



#!/usr/bin/env python
import doctest
import os
import sys

sys.path.insert(
    0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))


def test():
    for doc in ['SQLObject.rst']:
        doctest.testfile(doc, optionflags=doctest.ELLIPSIS)

if __name__ == '__main__':
    test()



import pytest
from sqlobject import DatabaseIndex, ForeignKey, IntCol, MultipleJoin, \
    SQLObject, StringCol
from sqlobject.dberrors import DatabaseError, IntegrityError, \
    OperationalError, ProgrammingError
from sqlobject.tests.dbtest import raises, setupClass, supports


########################################
# Indexes
########################################


class SOIndex1(SQLObject):
    name = StringCol(length=100)
    number = IntCol()

    nameIndex = DatabaseIndex('name', unique=True)
    nameIndex2 = DatabaseIndex(name, number)
    nameIndex3 = DatabaseIndex({'column': name,
                                'length': 3})


class SOIndex2(SQLObject):
    name = StringCol(length=100)
    nameIndex = DatabaseIndex({'expression': 'lower(name)'})


def test_indexes_1():
    setupClass(SOIndex1)
    n = 0
    for name in 'blah blech boring yep yort snort'.split():
        n += 1
        SOIndex1(name=name, number=n)
    mod = SOIndex1._connection.module
    raises(
        (mod.ProgrammingError, mod.IntegrityError,
         mod.OperationalError, mod.DatabaseError,
         ProgrammingError, IntegrityError, OperationalError, DatabaseError),
        SOIndex1, name='blah', number=0)


def test_indexes_2():
    if not supports('expressionIndex'):
        pytest.skip("expressionIndex isn't supported")
    setupClass(SOIndex2)
    SOIndex2(name='')


class PersonIndexGet(SQLObject):
    firstName = StringCol(length=100)
    lastName = StringCol(length=100)
    age = IntCol(alternateID=True)
    nameIndex = DatabaseIndex(firstName, lastName, unique=True)


def test_index_get_1():
    setupClass(PersonIndexGet, force=True)

    PersonIndexGet(firstName='Eric', lastName='Idle', age=62)
    PersonIndexGet(firstName='Terry', lastName='Gilliam', age=65)
    PersonIndexGet(firstName='John', lastName='Cleese', age=66)

    PersonIndexGet.get(1)
    PersonIndexGet.nameIndex.get('Terry', 'Gilliam')
    PersonIndexGet.nameIndex.get(firstName='John', lastName='Cleese')

    raises(Exception, PersonIndexGet.nameIndex.get,
           firstName='Graham', lastName='Chapman')

    raises(Exception, PersonIndexGet.nameIndex.get,
           'Terry', lastName='Gilliam')

    raises(Exception, PersonIndexGet.nameIndex.get, 'Terry', 'Gilliam', 65)

    raises(Exception, PersonIndexGet.nameIndex.get, 'Terry')


class PersonIndexGet2(SQLObject):
    name = StringCol(alternateID=True, length=100)
    age = IntCol()
    addresses = MultipleJoin('AddressIndexGet2')


class AddressIndexGet2(SQLObject):
    person = ForeignKey('PersonIndexGet2', notNone=True)
    type = StringCol(notNone=True, length=100)
    street = StringCol(notNone=True)
    pk = DatabaseIndex(person, type, unique=True)


def test_index_get_2():
    setupClass([PersonIndexGet2, AddressIndexGet2])

    p = PersonIndexGet2(name='Terry Guilliam', age=64)
    AddressIndexGet2(person=p, type='home', street='Terry Street 234')
    AddressIndexGet2(person=p, type='work', street='Guilliam Street 234')

    AddressIndexGet2.pk.get(p, 'work')
    AddressIndexGet2.pk.get(person=p, type='work')



"""
This is a not-very-formal outline of the interface that SQLObject
provides.  While its in the form of a formal interface, it doesn't
use any interface system.
"""


class Interface(object):
    pass


class ISQLObject(Interface):

    sqlmeta = """
    A class or instance representing internal state and methods for
    introspecting this class.

    ``MyClass.sqlmeta`` is a class, and ``myInstance.sqlmeta`` is an
    instance of this class.  So every instance gets its own instance
    of the metadata.

    This object follows the ``Isqlmeta`` interface.
    """

    # classmethod
    def get(id, connection=None):
        """
        Returns the object with the given `id`.  If `connection` is
        given, then get the object from the given connection
        (otherwise use the default or configured connection)

        It raises ``SQLObjectNotFound`` if no row exists with that ID.
        """

    # classmethod
    def selectBy(connection=None, **attrs):
        """
        Performs a ``SELECT`` in the given `connection` (or default
        connection).

        Each of the keyword arguments should be a column, and the
        equality comparisons will be ``ANDed`` together to produce the
        result.
        """

    # classmethod
    def dropTable(ifExists=False, dropJoinTables=True, cascade=False,
                  connection=None):
        """
        Drops this table from the database.  If ``ifExists`` is true,
        then it is not an error if the table doesn't exist.

        Join tables (mapping tables for many-to-many joins) are
        dropped if this class comes alphabetically before the other
        join class, and if ``dropJoinTables`` is true.

        ``cascade`` is passed to the connection, and if true tries to
        drop tables that depend on this table.
        """

    # classmethod
    def createTable(ifNotExists=False, createJoinTables=True,
                    createIndexes=True, connection=None):
        """
        Creates the table.  If ``ifNotExists`` is true, then it is not
        an error if the table already exists.

        Join tables (mapping tables for many-to-many joins) are
        created if this class comes alphabetically before the other
        join class, and if ``createJoinTables`` is true.

        If ``createIndexes`` is true, indexes are also created.
        """

    # classmethod
    def createTableSQL(createJoinTables=True, connection=None,
                       createIndexes=True):
        """
        Returns the SQL that would be sent with the analogous call
        to ``Class.createTable(...)``
        """

    def sync():
        """
        This will refetch the data from the database, putting it in
        sync with the database (in case another process has modified
        the database since this object was fetched).  It will raise
        ``SQLObjectNotFound`` if the row has been deleted.

        This will call ``self.syncUpdate()`` if ``lazyUpdates`` are
        on.
        """

    def syncUpdate():
        """
        If ``.sqlmeta.lazyUpdates`` is true, then this method must be
        called to push accumulated updates to the server.
        """

    def expire():
        """
        This will remove all the column information from the object.
        The next time this information is used, a ``SELECT`` will be
        made to re-fetch the data.  This is like a lazy ``.sync()``.
        """

    def set(**attrs):
        """
        This sets many column attributes at once.  ``obj.set(a=1,
        b=2)`` is equivalent to ``obj.a=1; obj.b=2``, except that it
        will be grouped into one ``UPDATE``
        """

    def destroySelf():
        """
        Deletes this row from the database.  This is called on
        instances, not on the class.  The object still persists,
        because objects cannot be deleted from the Python process
        (they can only be forgotten about, at which time they are
        garbage collected).  The object becomes obsolete, and further
        activity on it will raise errors.
        """

    def sqlrepr(obj, connection=None):
        """
        Returns the SQL representation of the given object, for the
        configured database connection.
        """


class Isqlmeta(Interface):

    table = """
    The name of the table in the database.  This is derived from
    ``style`` and the class name if no explicit name is given.
    """

    idName = """
    The name of the primary key column in the database.  This is
    derived from ``style`` if no explicit name is given.
    """

    idType = """
    A function that coerces/normalizes IDs when setting IDs.  This
    is ``int`` by default (all IDs are normalized to integers).
    """

    style = """
    An instance of the ``IStyle`` interface.  This maps Python
    identifiers to database names.
    """

    lazyUpdate = """
    A boolean (default false).  If true, then setting attributes on
    instances (or using ``inst.set(...)`` will not send ``UPDATE``
    queries immediately (you must call ``inst.syncUpdates()`` or
    ``inst.sync()`` first).
    """

    defaultOrder = """
    When selecting objects and not giving an explicit order, this
    attribute indicates the default ordering.  It is like this value
    is passed to ``.select()`` and related methods; see those method's
    documentation for details.
    """

    cacheValues = """
    A boolean (default true).  If true, then the values in the row are
    cached as long as the instance is kept (and ``inst.expire()`` is
    not called).  If false, then every attribute access causes a
    ``SELECT`` (which is absurdly inefficient).
    """

    registry = """
    Because SQLObject uses strings to relate classes, and these
    strings do not respect module names, name clashes will occur if
    you put different systems together.  This string value serves
    as a namespace for classes.
    """

    fromDatabase = """
    A boolean (default false).  If true, then on class creation the
    database will be queried for the table's columns, and any missing
    columns (possible all columns) will be added automatically.
    """

    columns = """
    A dictionary of ``{columnName: anSOColInstance}``.  You can get
    information on the columns via this read-only attribute.
    """

    columnList = """
    A list of the values in ``columns``.  Sometimes a stable, ordered
    version of the columns is necessary; this is used for that.
    """

    columnDefinitions = """
    A dictionary like ``columns``, but contains the original column
    definitions (which are not class-specific, and have no logic).
    """

    joins = """
    A list of all the Join objects for this class.
    """

    indexes = """
    A list of all the indexes for this class.
    """

    # Instance attributes

    expired = """
    A boolean.  If true, then the next time this object's column
    attributes are accessed a query will be run.
    """

    # Methods

    def addColumn(columnDef, changeSchema=False, connection=None):
        """
        Adds the described column to the table.  If ``changeSchema``
        is true, then an ``ALTER TABLE`` query is called to change the
        database.

        Attributes given in the body of the SQLObject subclass are
        collected and become calls to this method.
        """

    def delColumn(column, changeSchema=False, connection=None):
        """
        Removes the given column (either the definition from
        ``columnDefinition`` or the SOCol object from ``columns``).

        If ``changeSchema`` is true, then an ``ALTER TABLE`` query is
        made.
        """

    def addColumnsFromDatabase(connection=None):
        """
        Adds all the columns from the database that are not already
        defined.  If the ``fromDatabase`` attribute is true, then
        this is called on class instantiation.
        """

    def addJoin(joinDef):
        """
        Adds a join to the class.
        """

    def delJoin(joinDef):
        """
        Removes a join from the class.
        """

    def addIndex(indexDef):
        """
        Adds the index to the class.
        """

    def asDict():
        """
        Returns the SQLObject instance as a dictionary (column names
        as keys, column values as values).

        Use like::

            ASQLObjectClass(a=1, b=2).asDict()

        Which should return ``{'a': 1, 'b': 2}``.

        Note: this is a *copy* of the object's columns; changing the
        dictionary will not effect the object it was created from.
        """


class ICol(Interface):

    def __init__(name=None, **kw):
        """
        Creates a column definition.  This is an object that describes
        a column, basically just holding the keywords for later
        creating an ``SOCol`` (or subclass) instance.  Subclasses of
        ``Col`` (which implement this interface) typically create the
        related subclass of ``SOCol``.
        """

    name = """
    The name of the column.  If this is not given in the constructor,
    ``SQLObject`` will set this attribute from the variable name this
    object is assigned to.
    """


class ISOCol(Interface):

    """
    This is a column description that is bound to a single class.
    This cannot be shared by subclasses, so a new instance is created
    for every new class (in cases where classes share columns).

    These objects are created by ``Col`` instances, you do not create
    them directly.
    """

    name = """
    The name of the attribute that points to this column.  This is the
    Python name of the column.
    """

    columnDef = """
    The ``Col`` object that created this instance.
    """

    immutable = """
    Boolean, default false.  If true, then this column cannot be
    modified.  It cannot even be modified at construction, rendering
    the table read-only.  This will probably change in the future, as
    it renders the option rather useless.
    """

    cascade = """
    If a foreign key, then this indicates if deletes in that foreign
    table should cascade into this table.  This can be true (deletes
    cascade), false (the default, they do not cascade), or ``'null'``
    (this column is set to ``NULL`` if the foreign key is deleted).
    """

    constraints = """
    A list of ... @@?
    """

    notNone = """
    Boolean, default false.  It true, then ``None`` (aka ``NULL``) is
    not allowed in this column.  Also the ``notNull`` attribute can be
    used.
    """

    foreignKey = """
    If not None, then this column points to another table.  The
    attribute is the name (a string) of that table/class.
    """

    dbName = """
    The name of this column in the database.
    """

    alternateID = """
    Boolean, default false.  If true, then this column is assumed to
    be unique, and you can fetch individual rows based on this
    column's value.  This will add a method ``byAttributeName`` to the
    parent SQLObject subclass.
    """

    unique = """
    Boolean, default false.  If this column is unique; effects the
    database ``CREATE`` statement, and is implied by
    ``alternateID=True``.
    """

    validator = """
    A IValidator object.  All setting of this column goes through the
    ``fromPython`` method of the validator.  All getting of this
    column from the database goes through ``toPython``.
    """

    default = """
    A value that holds the default value for this column.  If the
    default value passed in is a callable, then that value is called
    to return a default (a typical example being ``DateTime.now``).
    """

    sqlType = """
    The SQL type of the column, overriding the default column type.
    """



import pytest
from sqlobject import SQLObject
from sqlobject.tests.dbtest import setupClass, supports


class EmptyClass(SQLObject):

    pass


def test_empty():
    if not supports('emptyTable'):
        pytest.skip("emptyTable isn't supported")
    setupClass(EmptyClass)
    e1 = EmptyClass()
    e2 = EmptyClass()
    assert e1 != e2
    assert e1.id != e2.id
    assert e1 in list(EmptyClass.select())
    assert e2 in list(EmptyClass.select())
    e1.destroySelf()
    assert list(EmptyClass.select()) == [e2]



#!/usr/bin/env python

from os.path import abspath, dirname, join
from setuptools import setup
import sys

versionpath = join(abspath(dirname(__file__)), 'sqlobject', '__version__.py')
sqlobject_version = {}

if sys.version_info[:2] == (2, 7):
    execfile(versionpath, sqlobject_version)  # noqa: F821 'execfile' Py3

elif sys.version_info >= (3, 4):
    exec(open(versionpath, 'rU').read(), sqlobject_version)

else:
    raise ImportError("SQLObject requires Python 2.7 or 3.4+")

subpackages = ['firebird', 'include', 'include.tests',
               'inheritance', 'inheritance.tests',
               'manager', 'maxdb', 'mysql', 'mssql', 'postgres',
               'sqlite', 'sybase', 'tests', 'util',
               'versioning', 'versioning.test']

setup(
    name="SQLObject",
    version=sqlobject_version['version'],
    description="Object-Relational Manager, aka database wrapper",
    long_description="""\
SQLObject is a popular *Object Relational Manager* for providing an
object interface to your database, with tables as classes, rows as
instances, and columns as attributes.

SQLObject includes a Python-object-based query language that makes SQL
more abstract, and provides substantial database independence for
applications.

Supports MySQL, PostgreSQL, SQLite, Firebird, Sybase, MSSQL and MaxDB (SAPDB).
Python 2.7 or 3.4+ is required.

For development see the projects at
`SourceForge <https://sourceforge.net/projects/sqlobject/>`_
and `GitHub <https://github.com/sqlobject>`_.

.. image:: https://travis-ci.org/sqlobject/sqlobject.svg?branch=master
  :target: https://travis-ci.org/sqlobject/sqlobject
""",
    long_description_content_type="text/x-rst",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: "
        "GNU Library or Lesser General Public License (LGPL)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Database",
        "Topic :: Database :: Front-Ends",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    author="Ian Bicking",
    author_email="ianb@colorstudy.com",
    maintainer="Oleg Broytman",
    maintainer_email="phd@phdru.name",
    url="http://sqlobject.org/",
    download_url="https://pypi.org/project/SQLObject/%s/" %
    sqlobject_version['version'],
    project_urls={
        'Homepage': 'http://sqlobject.org/',
        'Development docs': 'http://sqlobject.org/devel/',
        'Download': 'https://pypi.org/project/SQLObject/%s/' %
        sqlobject_version['version'],
        'Github repo': 'https://github.com/sqlobject',
        'Issue tracker': 'https://github.com/sqlobject/sqlobject/issues',
        'SourceForge project': 'https://sourceforge.net/projects/sqlobject/',
        'Twitter': 'https://twitter.com/SQLObject',
        'Wikipedia': 'https://en.wikipedia.org/wiki/SQLObject',
    },
    keywords=["sql", "orm", "object-relational mapper"],
    license="LGPL",
    platforms="Any",
    packages=["sqlobject"]
    + ['sqlobject.%s' % package for package in subpackages],
    scripts=["scripts/sqlobject-admin", "scripts/sqlobject-convertOldURI"],
    package_data={
        "sqlobject.maxdb": ["readme.txt"],
    },
    entry_points="""
    [paste.filter_app_factory]
    main = sqlobject.wsgi_middleware:make_middleware
    """,
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
    requires=['FormEncode', 'PyDispatcher'],
    install_requires=[
        "FormEncode>=1.1.1,!=1.3.0; python_version=='2.7'",
        "FormEncode>=1.3.1; python_version>='3.4'",
        "PyDispatcher>=2.0.4",
    ],
    extras_require={
        # Firebird/Interbase
        'fdb': ['fdb'],
        'firebirdsql': ['firebirdsql'],
        'kinterbasdb': ['kinterbasdb'],
        # MS SQL
        'adodbapi': ['adodbapi'],
        'pymssql': ['pymssql'],
        # MySQL
        'mysql:python_version=="2.7"': ['MySQL-python'],
        'mysql:python_version>="3.4"': ['mysqlclient'],
        'mysql-connector': ['mysql-connector'],
        'oursql': ['oursql'],
        'pymysql': ['pymysql'],
        # ODBC
        'odbc': ['pyodbc'],
        'pyodbc': ['pyodbc'],
        'pypyodbc': ['pypyodbc'],
        # PostgreSQL
        'psycopg2': ['psycopg2'],
        'psycopg': ['psycopg2'],
        'postgres': ['psycopg2'],
        'postgresql': ['psycopg2'],
        'pygresql': ['pygresql'],
        'pypostgresql': ['py-postgresql'],
        'py-postgresql': ['py-postgresql'],
        'pg8000': ['pg8000'],
        #
        'sapdb': ['sapdb'],
        'sqlite': ['pysqlite'],
        'supersqlite': ['supersqlite'],
        'sybase': ['Sybase'],
    },
)

# Send announce to:
#   sqlobject-discuss@lists.sourceforge.net
#   python-announce@python.org
#   python-list@python.org
#   db-sig@python.org

# Email tempate:
"""
@@ INTRO


What's new in SQLObject
=======================

@@ CHANGES

For a more complete list, please see the news:
http://sqlobject.org/docs/News.html


What is SQLObject
=================

SQLObject is an object-relational mapper.  Your database tables are described
as classes, and rows are instances of those classes.  SQLObject is meant to be
easy to use and quick to get started with.

It currently supports MySQL through the `MySQLdb` package, PostgreSQL
through the `psycopg` package, SQLite, Firebird, MaxDB (SAP DB), MS SQL
and Sybase.  Python 2.7 or 3.4+ is required.


Where is SQLObject
==================

Site:
http://sqlobject.org

Mailing list:
https://lists.sourceforge.net/mailman/listinfo/sqlobject-discuss

Download:
https://pypi.org/project/SQLObject/@@/

News and changes:
http://sqlobject.org/docs/News.html

-- 
Ian Bicking  /  ianb@colorstudy.com  / http://blog.ianbicking.org
"""  # noqa: preserve space after two dashes



from __future__ import print_function

from sqlobject import ForeignKey, ManyToMany, OneToMany, SQLObject, StringCol
from sqlobject.tests.dbtest import setupClass


########################################
# Joins
########################################


class PersonJNew(SQLObject):

    name = StringCol(length=40, alternateID=True)
    addressJs = ManyToMany('AddressJNew')


class AddressJNew(SQLObject):

    zip = StringCol(length=5, alternateID=True)
    personJs = ManyToMany('PersonJNew')


class ImplicitJoiningSONew(SQLObject):
    foo = ManyToMany('Bar')


class ExplicitJoiningSONew(SQLObject):
    foo = OneToMany('Bar')


class TestJoin:

    def setup_method(self, meth):
        setupClass(PersonJNew)
        setupClass(AddressJNew)
        for n in ['bob', 'tim', 'jane', 'joe', 'fred', 'barb']:
            PersonJNew(name=n)
        for z in ['11111', '22222', '33333', '44444']:
            AddressJNew(zip=z)

    def test_join(self):
        b = PersonJNew.byName('bob')
        assert list(b.addressJs) == []
        z = AddressJNew.byZip('11111')
        b.addressJs.add(z)
        self.assertZipsEqual(b.addressJs, ['11111'])
        print(str(z.personJs), repr(z.personJs))
        self.assertNamesEqual(z.personJs, ['bob'])
        z2 = AddressJNew.byZip('22222')
        b.addressJs.add(z2)
        print(str(b.addressJs))
        self.assertZipsEqual(b.addressJs, ['11111', '22222'])
        self.assertNamesEqual(z2.personJs, ['bob'])
        b.addressJs.remove(z)
        self.assertZipsEqual(b.addressJs, ['22222'])
        self.assertNamesEqual(z.personJs, [])

    def assertZipsEqual(self, zips, dest):
        assert [a.zip for a in zips] == dest

    def assertNamesEqual(self, people, dest):
        assert [p.name for p in people] == dest

    def test_joinAttributeWithUnderscores(self):
        # Make sure that the implicit setting of joinMethodName works
        assert hasattr(ImplicitJoiningSONew, 'foo')
        assert not hasattr(ImplicitJoiningSONew, 'bars')

        # And make sure explicit setting also works
        assert hasattr(ExplicitJoiningSONew, 'foo')
        assert not hasattr(ExplicitJoiningSONew, 'bars')


class PersonJNew2(SQLObject):

    name = StringCol('name', length=40, alternateID=True)
    addressJ2s = OneToMany('AddressJNew2')


class AddressJNew2(SQLObject):

    class sqlmeta:
        defaultOrder = ['-zip', 'plus4']

    zip = StringCol(length=5)
    plus4 = StringCol(length=4, default=None)
    personJNew2 = ForeignKey('PersonJNew2')


class TestJoin2:

    def setup_method(self, meth):
        setupClass([PersonJNew2, AddressJNew2])
        p1 = PersonJNew2(name='bob')
        p2 = PersonJNew2(name='sally')
        for z in ['11111', '22222', '33333']:
            AddressJNew2(zip=z, personJNew2=p1)
        AddressJNew2(zip='00000', personJNew2=p2)

    def test_basic(self):
        bob = PersonJNew2.byName('bob')
        sally = PersonJNew2.byName('sally')
        print(bob.addressJ2s)
        print(bob)
        assert len(list(bob.addressJ2s)) == 3
        assert len(list(sally.addressJ2s)) == 1
        bob.addressJ2s[0].destroySelf()
        assert len(list(bob.addressJ2s)) == 2
        z = bob.addressJ2s[0]
        z.zip = 'xxxxx'
        id = z.id
        del z
        z = AddressJNew2.get(id)
        assert z.zip == 'xxxxx'

    def test_defaultOrder(self):
        p1 = PersonJNew2.byName('bob')
        assert ([i.zip for i in p1.addressJ2s]
                == ['33333', '22222', '11111'])


_personJ3_getters = []
_personJ3_setters = []


class PersonJNew3(SQLObject):

    name = StringCol('name', length=40, alternateID=True)
    addressJNew3s = OneToMany('AddressJNew3')


class AddressJNew3(SQLObject):

    zip = StringCol(length=5)
    personJNew3 = ForeignKey('PersonJNew3')

    def _get_personJNew3(self):
        value = self._SO_get_personJNew3()
        _personJ3_getters.append((self, value))
        return value

    def _set_personJNew3(self, value):
        self._SO_set_personJNew3(value)
        _personJ3_setters.append((self, value))


class TestJoin3:

    def setup_method(self, meth):
        setupClass([PersonJNew3, AddressJNew3])
        p1 = PersonJNew3(name='bob')
        p2 = PersonJNew3(name='sally')
        for z in ['11111', '22222', '33333']:
            AddressJNew3(zip=z, personJNew3=p1)
        AddressJNew3(zip='00000', personJNew3=p2)

    def test_accessors(self):
        assert len(list(_personJ3_getters)) == 0
        assert len(list(_personJ3_setters)) == 4
        bob = PersonJNew3.byName('bob')
        for addressJ3 in bob.addressJNew3s:
            addressJ3.personJNew3
        assert len(list(_personJ3_getters)) == 3
        assert len(list(_personJ3_setters)) == 4



import pickle
import pytest
from sqlobject import IntCol, SQLObject, StringCol
from sqlobject.tests.dbtest import getConnection, raises, setupClass


########################################
# Pickle instances
########################################


class SOTestPickle(SQLObject):
    question = StringCol()
    answer = IntCol()


test_question = 'The Ulimate Question of Life, the Universe and Everything'
test_answer = 42


def test_pickleCol():
    setupClass(SOTestPickle)
    connection = SOTestPickle._connection
    test = SOTestPickle(question=test_question, answer=test_answer)

    pickle_data = pickle.dumps(test, pickle.HIGHEST_PROTOCOL)
    connection.cache.clear()
    test = pickle.loads(pickle_data)
    test2 = connection.cache.tryGet(test.id, SOTestPickle)

    assert test2 is test
    assert test.question == test_question
    assert test.answer == test_answer

    if (connection.dbName == 'sqlite') and connection._memory:
        pytest.skip("The following test requires a different connection")

    test = SOTestPickle.get(
        test.id,
        # make a different DB URI and open another connection
        connection=getConnection(registry=''))
    raises(pickle.PicklingError, pickle.dumps, test, pickle.HIGHEST_PROTOCOL)



from sqlobject import SQLObject, StringCol, sqlmeta
from sqlobject.tests.dbtest import setupClass

real_sqlmeta = sqlmeta


class Reparented1(SQLObject):

    class sqlmeta:
        table = 'reparented1'

    dummy = StringCol()


class Reparented2(SQLObject):
    class sqlmeta(object):
        @classmethod
        def setClass(cls, soClass):
            # Well, it's pretty hard to call the superclass method
            # when it's a classmethod and it's not actually your
            # *current* superclass.  Sigh
            real_sqlmeta.setClass.__func__(cls, soClass)
            cls.worked = True

    dummy = StringCol()


def test_reparented():
    setupClass([Reparented1, Reparented2])
    assert Reparented1.sqlmeta.table == 'reparented1'
    assert issubclass(Reparented1.sqlmeta, real_sqlmeta)
    assert issubclass(Reparented2.sqlmeta, real_sqlmeta)
    assert Reparented2.sqlmeta.worked



import pytest
from sqlobject import SQLObject, StringCol
from sqlobject.dberrors import DuplicateEntryError, ProgrammingError
from sqlobject.tests.dbtest import getConnection, raises, setupClass, supports


########################################
# Table aliases and self-joins
########################################


class SOTestException(SQLObject):
    name = StringCol(unique=True, length=100)


class SOTestExceptionWithNonexistingTable(SQLObject):
    pass


def test_exceptions():
    if not supports("exceptions"):
        pytest.skip("exceptions aren't supported")
    setupClass(SOTestException)
    SOTestException(name="test")
    raises(DuplicateEntryError, SOTestException, name="test")

    connection = getConnection()
    if connection.module.__name__ != 'psycopg2':
        return
    SOTestExceptionWithNonexistingTable.setConnection(connection)
    try:
        list(SOTestExceptionWithNonexistingTable.select())
    except ProgrammingError as e:
        assert e.args[0].code == '42P01'
    else:
        assert False, "DID NOT RAISE"
